@@ -1,17 +1,11 @@
 # Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
 # MIT License (see LICENSE or https://opensource.org/licenses/MIT)
 
-from __future__ import (absolute_import, division, print_function)
-__metaclass__ = type  # noqa (fixes E402 for the imports below)
-
 import collections
-
-from spnego._compat import (
-    IntEnum,
-)
+import enum
 
 
-class BufferType(IntEnum):
+class BufferType(enum.IntEnum):
     """Buffer types to use for an IOVBuffer type.
 
     These are the IOVBuffer type flags that can be set for an IOVBuffer. The keys are a generified name for the
